/*
 * ad7927.c
 *
 * The ad7927 is a 7 channel ADC from Analog Devices
 * Copyright (C) 2009 Andre Renaud <andre@bluewatersys.com>
 *
 * The AD7927 communicates with a host processor via an SPI/Microwire Bus
 * interface. 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/sysfs.h>
#include <linux/spi/spi.h>
#include <linux/spi/ad7927.h>

enum {
        AD7927_WRITE 	= 1 << 11,
        AD7927_SEQ 	= 1 << 10,
        AD7927_ADD2	= 1 << 8,
        AD7927_ADD1 	= 1 << 7,
        AD7927_ADD0 	= 1 << 6,
        AD7927_PM1 	= 1 << 5,
        AD7927_PM0 	= 1 << 4,
        AD7927_SHADOW	= 1 << 3,
        AD7927_DONTC	= 1 << 2,
        AD7927_RANGE	= 1 << 1,
        AD7927_CODING	= 1 << 0,
};

static int ad7927_set_control (struct spi_device *spi, int control)
{
        u8 tx_buf[2];
        u8 rx_buf[2];
        int status;
        struct spi_transfer t;
        struct spi_message message;

        /* Fix to align with datasheet numbering (12-bit register) */
        control = control << 4;
        tx_buf[0] = control >> 8;
        tx_buf[1] = control & 0xff;
        spi_message_init (&message);
        t.tx_buf = tx_buf;
        t.rx_buf = rx_buf;
        t.len = 2;
        spi_message_add_tail (&t, &message);
        status = spi_sync (spi, &message);

        if (status < 0) {
                dev_warn (&spi->dev, "spi xfer failed with status %d\n",
                                status);
                return status;
        }

	printk(KERN_DEBUG "ad7927: raw = %.4x\n", 
	       (rx_buf[0] << 8) | rx_buf[1]);
        return (rx_buf[0] << 8) | rx_buf[1];
}

static int ad7927_set_channel(struct spi_device *spi, int channel)
{
        int res_channel, res;
	u16 control;

        if (channel < 0 || channel > 7) {
                dev_err (&spi->dev, "Invalid ADC: %d\n", channel);
                return -EINVAL;
        }

	control = (AD7927_WRITE | AD7927_PM1 | AD7927_PM0 | AD7927_CODING | 
		   (channel << 6));
	
        ad7927_set_control(spi, control);        
        res = ad7927_set_control(spi, control);
	res_channel = (res >> 12) & 0x7;
	if (res_channel != channel) {
		dev_err(&spi->dev, "Channel mismatch. Wrote %d, read %d\n",
			channel, res_channel);
		return -EFAULT;
	}

	return res;
}

#define CHANNEL(x) \
static ssize_t ad7927_get_raw_##x (struct device *dev,              \
                struct device_attribute *devattr, char *buf) {          \
        struct spi_device *spi = to_spi_device(dev);                    \
        int res = ad7927_set_channel(spi, x);                           \
        return sprintf (buf, "0x%4.4x\n", res & 0xfff);                 \
}                                                                       \
static ssize_t ad7927_get_millivolts_##x (struct device *dev,           \
                struct device_attribute *devattr, char *buf) {          \
        struct spi_device *spi = to_spi_device(dev);                    \
        struct ad7927_platform_data *pdata = dev->platform_data;        \
        int ref_in;                                                     \
        int res = ad7927_set_channel(spi, x);                           \
        ref_in = pdata ? pdata->ref_in : 2500;                          \
        res = (res & 0xfff) * ref_in * 2 / 4095;                        \
        return sprintf (buf, "%d\n", res);                              \
}                                                                       \
static DEVICE_ATTR(raw##x, S_IRUGO, ad7927_get_raw_##x, NULL);  \
static DEVICE_ATTR(millivolts##x, S_IRUGO, ad7927_get_millivolts_##x, NULL);

CHANNEL(0);
CHANNEL(1);
CHANNEL(2);
CHANNEL(3);
CHANNEL(4);
CHANNEL(5);
CHANNEL(6);
CHANNEL(7);

static ssize_t ad7927_get_ref_in(struct device *dev, struct device_attribute
			      *devattr, char *buf)
{
        struct ad7927_platform_data *pdata = dev->platform_data;
        if (pdata)
                return sprintf (buf, "%d\n", pdata->ref_in);
        else
                return sprintf (buf, "2500\n");
}

static DEVICE_ATTR(ref_in, S_IRUGO, ad7927_get_ref_in, NULL);

static struct attribute *ad7927_attrs[] = {
        &dev_attr_ref_in.attr,
        &dev_attr_raw0.attr,
        &dev_attr_millivolts0.attr,
        &dev_attr_raw1.attr,
        &dev_attr_millivolts1.attr,
        &dev_attr_raw2.attr,
        &dev_attr_millivolts2.attr,
        &dev_attr_raw3.attr,
        &dev_attr_millivolts3.attr,
        &dev_attr_raw4.attr,
        &dev_attr_millivolts4.attr,
        &dev_attr_raw5.attr,
        &dev_attr_millivolts5.attr,
        &dev_attr_raw6.attr,
        &dev_attr_millivolts6.attr,
        &dev_attr_raw7.attr,
        &dev_attr_millivolts7.attr,
        NULL,
};

static const struct attribute_group sysfs_files = {
        .attrs = ad7927_attrs,
};
/*----------------------------------------------------------------------*/

static int __devinit ad7927_probe(struct spi_device *spi)
{
        int status = sysfs_create_group (&spi->dev.kobj, &sysfs_files);
        if (status) {
		dev_dbg(&spi->dev, "device_create_file failure.\n");
		return status;
	}
        
        status = ad7927_set_control (spi, AD7927_WRITE | AD7927_PM1 | 
                                        AD7927_PM0 | AD7927_CODING);
        if (status < 0) {
		dev_err(&spi->dev, "probe status = 0x%x\n", status);
                //sysfs_remove_group (&spi->dev.kobj, &sysfs_files);
                //return status;
        }

        if (!spi->dev.platform_data)
                dev_warn (&spi->dev, "no platform data - guessing ref_in of 2500mV\n");

        return 0;
}


static int __devexit ad7927_remove(struct spi_device *spi)
{
        sysfs_remove_group(&spi->dev.kobj, &sysfs_files);
	return 0;
}

static struct spi_driver ad7927_driver = {
	.driver = {
		.name	= "ad7927",
		.owner	= THIS_MODULE,
	},
	.probe	= ad7927_probe,
	.remove	= __devexit_p(ad7927_remove),
};

static int __init init_ad7927(void)
{
	return spi_register_driver(&ad7927_driver);
}

static void __exit cleanup_ad7927(void)
{
	spi_unregister_driver(&ad7927_driver);
}

module_init(init_ad7927);
module_exit(cleanup_ad7927);

MODULE_AUTHOR("Andre Renaud");
MODULE_DESCRIPTION("AD7927 ADC Driver");
MODULE_LICENSE("GPL");
