/*
 *  drivers/misc/ad524x.c
 *
 *  Author:     Theuns Verwoerd <theuns@bluewatersys.com>
 *  Created:    Apr 20, 2009
 *  Copyright:  Bluewater Systems Ltd
 *
 * I2C driver for Analog Devices AD5243/AD5428 Digital Potentiometer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/i2c.h>

enum functions { ad5243, ad5248 };
enum flags { CHANNEL_1 = 0, CHANNEL_2 = 0x80, DISABLE_FLAG = 0x40 };

// Currently ad5248 is only tested on address 0x2f
static const struct i2c_device_id ad524x_id[] = {
	{ "ad5243", ad5243 },
	{ "ad5248", ad5248 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, ad524x_id);

struct ad524x {
	struct i2c_client *client;
	const struct i2c_device_id *id;
	char	flags;		// make disabled flag persistent across ops
	struct mutex update_lock;
};

static struct i2c_driver ad524x_driver;

/* Sysfs access routines */

static ssize_t show_channel(struct device *dev, struct device_attribute *attr, char *buf, int channel)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ad524x *ad524x = i2c_get_clientdata(client);
	int regVal = ~0;

	mutex_lock(&ad524x->update_lock);
	/* Write channel selection */
	regVal = i2c_smbus_write_byte(client, channel | ad524x->flags);
	/* Read back data */
	regVal = i2c_smbus_read_byte(client);
	mutex_unlock(&ad524x->update_lock);

	return sprintf(buf, "0x%02x\n", regVal);
}

static ssize_t set_channel(struct device *dev, struct device_attribute *attr, const char *buf, size_t count, int channel)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ad524x *ad524x = i2c_get_clientdata(client);
	int regVal = simple_strtoul(buf, NULL, 0);

	mutex_lock(&ad524x->update_lock);
	i2c_smbus_write_byte_data(client, channel | ad524x->flags, regVal);
	mutex_unlock(&ad524x->update_lock);
	return count;
}


static ssize_t show_channel_1(struct device *dev, struct device_attribute *attr, char *buf)
{
	return show_channel(dev, attr, buf, CHANNEL_1);
}
static ssize_t set_channel_1(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	return set_channel(dev, attr, buf, count, CHANNEL_1);
}
static DEVICE_ATTR(channel_1, S_IRUGO | S_IWUSR, show_channel_1, set_channel_1);


static ssize_t show_channel_2(struct device *dev, struct device_attribute *attr, char *buf)
{
	return show_channel(dev, attr, buf, CHANNEL_2);
}
static ssize_t set_channel_2(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	return set_channel(dev, attr, buf, count, CHANNEL_2);
}
static DEVICE_ATTR(channel_2, S_IRUGO | S_IWUSR, show_channel_2, set_channel_2);


static ssize_t show_disabled(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ad524x *ad524x = i2c_get_clientdata(client);
	
 	return sprintf(buf, "%i\n", ad524x->flags ? 1 : 0);
}
static ssize_t set_disabled(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct ad524x *ad524x = i2c_get_clientdata(client);
	int regVal = simple_strtoul(buf, NULL, 0);

	mutex_lock(&ad524x->update_lock);
	// This will change read channel to 0 implicitly, but we don't rely on that anyway.
	ad524x->flags = (regVal ? DISABLE_FLAG : 0);
	i2c_smbus_write_byte(client, ad524x->flags);
	mutex_unlock(&ad524x->update_lock);
	return count;
}
static DEVICE_ATTR(disabled, S_IRUGO | S_IWUSR, show_disabled, set_disabled);

static struct attribute *ad524x_attrs[] = {
	&dev_attr_channel_1.attr,
	&dev_attr_channel_2.attr,
	&dev_attr_disabled.attr,
	NULL,
};
static const struct attribute_group at524x_sysfs_files = {
	.attrs = ad524x_attrs,
};

/* Initialisation/framework code */

static int ad524x_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	struct ad524x *ad524x;
	int err = 0;

	dev_info(&client->dev, "addr %x type %lx\n", client->addr, id->driver_data);
	
	ad524x = kzalloc(sizeof(struct ad524x), GFP_KERNEL);
	if (!ad524x)
		return -ENOMEM;

	ad524x->client = client;
	ad524x->id = id;
	ad524x->flags = 0;
	mutex_init(&ad524x->update_lock);
	i2c_set_clientdata(client, ad524x);
	
	/* Register sysfs files */
	err = sysfs_create_group(&client->dev.kobj, &at524x_sysfs_files);
	if (err) {
		dev_dbg(&client->dev, "sysfs_create_group.\n");
		goto out_free;
	}
	
	return 0;
	
out_free:
	i2c_set_clientdata(client, NULL);
	kfree(ad524x);
	return err;
}

static int __devexit ad524x_remove(struct i2c_client *client)
{
	struct ad524x *ad524x = i2c_get_clientdata(client);

	printk("\t\t%s:%u\n", __FUNCTION__, __LINE__);

	sysfs_remove_group(&client->dev.kobj, &at524x_sysfs_files);
	i2c_set_clientdata(client, NULL);
	kfree(ad524x);
	return 0;
}

#ifdef CONFIG_PM
static int ad524x_suspend(struct i2c_client *client, pm_message_t state)
{
	struct ad524x *ad524x = i2c_get_clientdata(client);

// 	if (client->irq >= 0 && device_may_wakeup(&client->dev))
// 		enable_irq_wake(client->irq);
	mutex_lock(&ad524x->update_lock);
	return 0;
}

static int ad524x_resume(struct i2c_client *client)
{
	struct ad524x *ad524x = i2c_get_clientdata(client);
	
// 	if (client->irq >= 0 && device_may_wakeup(&client->dev))
// 		disable_irq_wake(client->irq);
	mutex_unlock(&ad524x->update_lock);
	return 0;
}
#else
#define ad524x_suspend	NULL
#define ad524x_resume	NULL
#endif

static struct i2c_driver ad524x_driver = {
	.driver = {
		.name	= "ad524x",
		.owner	= THIS_MODULE,
	},
	.id_table = ad524x_id,
	
	.probe = ad524x_probe,
	.remove = __devexit_p(ad524x_remove),
	.suspend = ad524x_suspend,
	.resume = ad524x_resume,
};

static int __init ad524x_init(void)
{
	return i2c_add_driver(&ad524x_driver);
}

static void __exit ad524x_exit(void)
{
	i2c_del_driver(&ad524x_driver);
}

module_init(ad524x_init);
module_exit(ad524x_exit);

MODULE_DESCRIPTION("ad524x driver");
MODULE_AUTHOR("Theuns Verwoerd <theuns@bluewatersys.com>");
MODULE_LICENSE("GPL");
