/*
 *  drivers/misc/fm31l27x.c
 *
 *  Author:     Theuns Verwoerd <theuns@bluewatersys.com>
 *  Created:    Apr 07, 2009
 *  Copyright:  Bluewater Systems Ltd
 *
 * I2C driver for Ramtron FM31L27x Integrated Processor Companion with Memory chip.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/rtc.h>
#include <linux/bcd.h>

enum functions { fm31l27x_ram, fm31l27x_reg };

static const struct i2c_device_id fm31l27x_id[] = {
	{ "fm31l27x_ram", fm31l27x_ram },
	{ "fm31l27x_reg", fm31l27x_reg },
	{ }
};
MODULE_DEVICE_TABLE(i2c, fm31l27x_id);

struct fm31l27x {
	struct i2c_client *client;
	const struct i2c_device_id *id;
	struct rtc_device *rtc;

	struct mutex update_lock;
};

static struct i2c_driver fm31l27x_driver;

/* FM31L27x registers */
#define FM31L27X_SERIAL_7		0x18
#define FM31L27X_SERIAL_6		0x17
#define FM31L27X_SERIAL_5		0x16
#define FM31L27X_SERIAL_4		0x15
#define FM31L27X_SERIAL_3		0x14
#define FM31L27X_SERIAL_2		0x13
#define FM31L27X_SERIAL_1		0x12
#define FM31L27X_SERIAL_0		0x11
#define FM31L27X_CNT_2_MSB		0x10
#define FM31L27X_CNT_2_LSB		0x0F
#define FM31L27X_CNT_1_MSB		0x0E
#define FM31L27X_CNT_1_LSB		0x0D
#define FM31L27X_CNT_CTRL		0x0C
#define FM31L27X_COMP_CTRL		0x0B
#define FM31L27X_WDOG_CTRL		0x0A
#define FM31L27X_WDOG_FLAG		0x09
#define FM31L27X_YEAR			0x08
#define FM31L27X_MONTH			0x07
#define FM31L27X_DATE			0x06
#define FM31L27X_DAY			0x05
#define FM31L27X_HOUR			0x04
#define FM31L27X_MIN			0x03
#define FM31L27X_SEC			0x02
#define FM31L27X_CAL_CTRL		0x01
#define FM31L27X_RTC_CTRL		0x00

/*** sysfs functions ***********************/

static ssize_t show_raw_real(struct device *dev, struct device_attribute *attr, char *buf, int regNum)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);
	int regVal = ~0;

	mutex_lock(&fm31l27x->update_lock);
	regVal = i2c_smbus_read_byte_data(client, regNum);
	mutex_unlock(&fm31l27x->update_lock);

	return sprintf(buf, "0x%x\n", regVal);
}

#define show_raw(reg, regDef)	\
static ssize_t show_raw_##reg(struct device *dev, struct device_attribute *attr, char *buf)	\
{																								\
	return show_raw_real(dev, attr, buf, regDef);												\
}

/* Manually decoded display routines */
static ssize_t show_serial(struct device *dev, struct device_attribute *attr, char *buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);
	int regVal = ~0, i, ret = 0;

	mutex_lock(&fm31l27x->update_lock);
	ret = 0;
	for (i = FM31L27X_SERIAL_7; i >= FM31L27X_SERIAL_0; i--) {
		regVal = i2c_smbus_read_byte_data(client, i);
		ret += sprintf(buf + ret, "%02x", regVal);
	}
	mutex_unlock(&fm31l27x->update_lock);

	ret += sprintf(buf + ret, "\n");
	return ret;
}

static ssize_t show_count_2 (struct device *dev, struct device_attribute *attr, char *buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);
	unsigned int count;

	// TODO handle linked counter
	
	mutex_lock(&fm31l27x->update_lock);
	count = i2c_smbus_read_byte_data(client, FM31L27X_CNT_2_MSB) << 8 
			| i2c_smbus_read_byte_data(client, FM31L27X_CNT_2_LSB);
	mutex_unlock(&fm31l27x->update_lock);

	return sprintf(buf, "0x%x", count);
}

static ssize_t show_count_1 (struct device *dev, struct device_attribute *attr, char *buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);
	unsigned int count;

	mutex_lock(&fm31l27x->update_lock);
	count = i2c_smbus_read_byte_data(client, FM31L27X_CNT_1_MSB) << 8 
			| i2c_smbus_read_byte_data(client, FM31L27X_CNT_1_LSB);
	mutex_unlock(&fm31l27x->update_lock);

	return sprintf(buf, "0x%x", count);
}

static char* dayofweek[] = {0, "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};
static ssize_t show_date (struct device *dev, struct device_attribute *attr, char *buf)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);
	int year, month, date, day, hour, min, sec;

	mutex_lock(&fm31l27x->update_lock);
	/* First, latch clock reading in from hardware */
	i2c_smbus_write_byte_data(client, FM31L27X_RTC_CTRL, 0x1);
	
	/* Now read proper date */
	year = bcd2bin(i2c_smbus_read_byte_data(client, FM31L27X_YEAR));
	// Century overflow flag isn't writeable, so not much use.
	year += (year >= 70) ? 1900 : 2000;
	
	month = i2c_smbus_read_byte_data(client, FM31L27X_MONTH);
	date = i2c_smbus_read_byte_data(client, FM31L27X_DATE);
	day = i2c_smbus_read_byte_data(client, FM31L27X_DAY);
	hour = i2c_smbus_read_byte_data(client, FM31L27X_HOUR);
	min = i2c_smbus_read_byte_data(client, FM31L27X_MIN);
	sec = i2c_smbus_read_byte_data(client, FM31L27X_SEC);
	
	/* Clear out possible lock */
	i2c_smbus_write_byte_data(client, FM31L27X_RTC_CTRL, 0x0);
	mutex_unlock(&fm31l27x->update_lock);

	// Values are in BCD, so print them as hex
	return sprintf(buf, "%u-%02x-%02x %s %02x:%02x:%02x\n", year, month, date, dayofweek[day], hour, min, sec);
}

static ssize_t set_raw_real(struct device *dev, struct device_attribute *attr, const char *buf, size_t count, int regNum)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);
	int regVal = simple_strtoul(buf, NULL, 0);

	mutex_lock(&fm31l27x->update_lock);
	i2c_smbus_write_byte_data(client, regNum, regVal);
	mutex_unlock(&fm31l27x->update_lock);
	return count;
}

#define set_raw(reg, regDef)	\
static ssize_t set_raw_##reg(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)	\
{																		\
	return set_raw_real(dev, attr, buf, count, regDef);					\
}

#define set_show_etc(reg, regDef, w) 	\
	show_raw(reg, regDef);				\
	set_raw(reg, regDef);				\
	static DEVICE_ATTR(raw_##reg, S_IRUGO | (w ? S_IWUSR : 0), show_raw_##reg, set_raw_##reg);

set_show_etc(serial_7, FM31L27X_SERIAL_7, 1);
set_show_etc(serial_6, FM31L27X_SERIAL_6, 1);
set_show_etc(serial_5, FM31L27X_SERIAL_5, 1);
set_show_etc(serial_4, FM31L27X_SERIAL_4, 1);
set_show_etc(serial_3, FM31L27X_SERIAL_3, 1);
set_show_etc(serial_2, FM31L27X_SERIAL_2, 1);
set_show_etc(serial_1, FM31L27X_SERIAL_1, 1);
set_show_etc(serial_0, FM31L27X_SERIAL_0, 1);
set_show_etc(cnt_2_msb, FM31L27X_CNT_2_MSB, 1);
set_show_etc(cnt_2_lsb, FM31L27X_CNT_2_LSB, 1);
set_show_etc(cnt_1_msb, FM31L27X_CNT_1_MSB, 1);
set_show_etc(cnt_1_lsb, FM31L27X_CNT_1_LSB, 1);
set_show_etc(cnt_ctrl, FM31L27X_CNT_CTRL, 1);
set_show_etc(comp_ctrl, FM31L27X_COMP_CTRL, 1);
set_show_etc(wdog_ctrl, FM31L27X_WDOG_CTRL, 1);
set_show_etc(wdog_flag, FM31L27X_WDOG_FLAG, 1);
set_show_etc(year, FM31L27X_YEAR, 1);
set_show_etc(month, FM31L27X_MONTH, 1);
set_show_etc(date, FM31L27X_DATE, 1);
set_show_etc(day, FM31L27X_DAY, 1);
set_show_etc(hour, FM31L27X_HOUR, 1);
set_show_etc(min, FM31L27X_MIN, 1);
set_show_etc(sec, FM31L27X_SEC, 1);
set_show_etc(cal_ctrl, FM31L27X_CAL_CTRL, 1);
set_show_etc(rtc_ctrl, FM31L27X_RTC_CTRL, 1);

/* Decoded output files */
static DEVICE_ATTR(serial, S_IRUGO, show_serial, NULL);
static DEVICE_ATTR(count_1, S_IRUGO, show_count_1, NULL);
static DEVICE_ATTR(count_2, S_IRUGO, show_count_2, NULL);
static DEVICE_ATTR(date, S_IRUGO, show_date, NULL);

static struct attribute *fm31l28x_reg_attrs[] = {
	&dev_attr_raw_serial_7.attr,
	&dev_attr_raw_serial_6.attr,
	&dev_attr_raw_serial_5.attr,
	&dev_attr_raw_serial_4.attr,
	&dev_attr_raw_serial_3.attr,
	&dev_attr_raw_serial_2.attr,
	&dev_attr_raw_serial_1.attr,
	&dev_attr_raw_serial_0.attr,
	&dev_attr_raw_cnt_2_msb.attr,
	&dev_attr_raw_cnt_2_lsb.attr,
	&dev_attr_raw_cnt_1_msb.attr,
	&dev_attr_raw_cnt_1_lsb.attr,
	&dev_attr_raw_cnt_ctrl.attr,
	&dev_attr_raw_comp_ctrl.attr,
	&dev_attr_raw_wdog_ctrl.attr,
	&dev_attr_raw_wdog_flag.attr,
	&dev_attr_raw_year.attr,
	&dev_attr_raw_month.attr,
	&dev_attr_raw_date.attr,
	&dev_attr_raw_day.attr,
	&dev_attr_raw_hour.attr,
	&dev_attr_raw_min.attr,
	&dev_attr_raw_sec.attr,
	&dev_attr_raw_cal_ctrl.attr,
	&dev_attr_raw_rtc_ctrl.attr,
 
	&dev_attr_serial.attr,
	&dev_attr_count_1.attr,
	&dev_attr_count_2.attr,
	&dev_attr_date.attr,
	NULL,
};
static const struct attribute_group fm31l28x_reg_sysfs_files = {
	.attrs = fm31l28x_reg_attrs,
};

/*** RTC functions ***********************/

static int fm31l28x_read_time(struct device *dev, struct rtc_time *time)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);
	
	mutex_lock(&fm31l27x->update_lock);
	/* First, latch clock reading in from hardware */
	i2c_smbus_write_byte_data(client, FM31L27X_RTC_CTRL, 0x1);
	
	/* Now read proper date */
	time->tm_year = bcd2bin(i2c_smbus_read_byte_data(client, FM31L27X_YEAR));
	time->tm_year += (time->tm_year >= 70) ? 0 : 100;
	
	/* Garbage time settings will be filtered out by higher layers */
	time->tm_mon = bcd2bin(i2c_smbus_read_byte_data(client, FM31L27X_MONTH)) - 1;	// 1-base in hardware
	time->tm_mday = bcd2bin(i2c_smbus_read_byte_data(client, FM31L27X_DATE));
	time->tm_wday = bcd2bin(i2c_smbus_read_byte_data(client, FM31L27X_DAY));
	time->tm_hour = bcd2bin(i2c_smbus_read_byte_data(client, FM31L27X_HOUR));
	time->tm_min = bcd2bin(i2c_smbus_read_byte_data(client, FM31L27X_MIN));
	time->tm_sec = bcd2bin(i2c_smbus_read_byte_data(client, FM31L27X_SEC));
	
	time->tm_yday = -1;
	time->tm_isdst = -1;
	
	/* Clear out possible lock */
	i2c_smbus_write_byte_data(client, FM31L27X_RTC_CTRL, 0x0);
	mutex_unlock(&fm31l27x->update_lock);
	
	return 0;
}

static int fm31l28x_set_time(struct device *dev, struct rtc_time *time)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);

	mutex_lock(&fm31l27x->update_lock);
	/* First, latch clock for writing in to hardware */
	i2c_smbus_write_byte_data(client, FM31L27X_RTC_CTRL, 0x3);
	
	/* Now set proper date */
	i2c_smbus_write_byte_data(client, FM31L27X_YEAR, bin2bcd(time->tm_year % 100));
	
	i2c_smbus_write_byte_data(client, FM31L27X_MONTH, bin2bcd(time->tm_mon) + 1);	// 1-base in hardware
	i2c_smbus_write_byte_data(client, FM31L27X_DATE, bin2bcd(time->tm_mday));
	i2c_smbus_write_byte_data(client, FM31L27X_DAY, bin2bcd(time->tm_wday));
	i2c_smbus_write_byte_data(client, FM31L27X_HOUR, bin2bcd(time->tm_hour));
	i2c_smbus_write_byte_data(client, FM31L27X_MIN, bin2bcd(time->tm_min));
	i2c_smbus_write_byte_data(client, FM31L27X_SEC, bin2bcd(time->tm_sec));
	
	/* Flush */
	i2c_smbus_write_byte_data(client, FM31L27X_RTC_CTRL, 0x0);
	mutex_unlock(&fm31l27x->update_lock);
	
	return 0;
}

static const struct rtc_class_ops fm31l28x_rtc_ops = {
	.read_time = fm31l28x_read_time,
	.set_time = fm31l28x_set_time,
//	.read_alarm = ds1374_read_alarm,
//	.set_alarm = ds1374_set_alarm,
//	.ioctl = ds1374_ioctl,
};

/*** framework functions ***********************/

static int fm31l27x_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	struct fm31l27x *fm31l27x;
	int err = 0;

	dev_info(&client->dev, "addr %x type %lx\n", client->addr, id->driver_data);
	
	fm31l27x = kzalloc(sizeof(struct fm31l27x), GFP_KERNEL);
	if (!fm31l27x)
		return -ENOMEM;

	fm31l27x->client = client;
	fm31l27x->id = id;
	mutex_init(&fm31l27x->update_lock);
	i2c_set_clientdata(client, fm31l27x);
	
	/* Register sysfs files */
	switch (id->driver_data) {
		case fm31l27x_ram:
			break;
		case fm31l27x_reg:
			/* Switch on the charger */
			i2c_smbus_write_byte_data(client, FM31L27X_COMP_CTRL, 0x24);
			
			/* Switch on the clock oscillator */
			i2c_smbus_write_byte_data(client, FM31L27X_CAL_CTRL, 0x0);
			
			/* Register interfaces */
			fm31l27x->rtc = rtc_device_register(client->name, &client->dev,
											&fm31l28x_rtc_ops, THIS_MODULE);
			if (IS_ERR(fm31l27x->rtc)) {
				err = PTR_ERR(fm31l27x->rtc);
				dev_err(&client->dev, "unable to register the class device\n");
				goto out_free;
			}
			err = sysfs_create_group(&client->dev.kobj, &fm31l28x_reg_sysfs_files);
			if (err) {
				dev_dbg(&client->dev, "sysfs_create_group.\n");
				goto out_rtc;
			}
			break;
	}
	
	return 0;
	
out_rtc:
	rtc_device_unregister(fm31l27x->rtc);
out_free:
	i2c_set_clientdata(client, NULL);
	kfree(fm31l27x);
	return err;
}

static int __devexit fm31l27x_remove(struct i2c_client *client)
{
	struct fm31l27x *fm31l27x = i2c_get_clientdata(client);

	printk("\t\t%s:%u\n", __FUNCTION__, __LINE__);
	
	switch (fm31l27x->id->driver_data) {
		case fm31l27x_ram:
			break;
		case fm31l27x_reg:
			sysfs_remove_group(&client->dev.kobj, &fm31l28x_reg_sysfs_files);
			rtc_device_unregister(fm31l27x->rtc);
			break;
	}
		
	i2c_set_clientdata(client, NULL);
	kfree(fm31l27x);
	return 0;
}

#ifdef CONFIG_PM
static int fm31l27x_suspend(struct i2c_client *client, pm_message_t state)
{
// 	if (client->irq >= 0 && device_may_wakeup(&client->dev))
// 		enable_irq_wake(client->irq);
	return 0;
}

static int fm31l27x_resume(struct i2c_client *client)
{
// 	if (client->irq >= 0 && device_may_wakeup(&client->dev))
// 		disable_irq_wake(client->irq);
	return 0;
}
#else
#define fm31l27x_suspend	NULL
#define fm31l27x_resume	NULL
#endif

static struct i2c_driver fm31l27x_driver = {
	.driver = {
		.name	= "fm31l27x",
		.owner	= THIS_MODULE,
	},
	.id_table = fm31l27x_id,
	
	.probe = fm31l27x_probe,
	.remove = __devexit_p(fm31l27x_remove),
	.suspend = fm31l27x_suspend,
	.resume = fm31l27x_resume,
};

static int __init fm31l27x_init(void)
{
	return i2c_add_driver(&fm31l27x_driver);
}

static void __exit fm31l27x_exit(void)
{
	i2c_del_driver(&fm31l27x_driver);
}

module_init(fm31l27x_init);
module_exit(fm31l27x_exit);

MODULE_DESCRIPTION("FM31L27x driver");
MODULE_AUTHOR("Theuns Verwoerd <theuns@bluewatersys.com>");
MODULE_LICENSE("GPL");
